from abc import ABC
from dataclasses import dataclass
from typing import List, Dict

import torch.nn as nn
from torch import Tensor
from yacs.config import CfgNode

from fandak.core.datasets import GeneralBatch
from fandak.utils.torch import GeneralDataClass


@dataclass
class GeneralLoss(GeneralDataClass):
    """
    I assume that there is always the `main` attribute which is the loss that is
    going to be used for backpropagation.
    """

    main: Tensor


@dataclass
class GeneralForwardOut(GeneralDataClass):
    """
    The general output of the forward pass.
    """

    pass


class Model(nn.Module, ABC):
    def __init__(self, cfg: CfgNode):
        super().__init__()
        self.cfg = cfg

    def get_params(self, original_lr: float) -> List[Dict]:
        params = [{"params": self.parameters(), "lr": original_lr}]
        return params

    # noinspection PyMethodMayBeStatic
    def get_backprop_loss(self, loss: GeneralLoss) -> Tensor:
        return loss.main

    def forward(self, batch: GeneralBatch) -> GeneralForwardOut:
        raise NotImplementedError

    def loss(self, batch: GeneralBatch, forward_out: GeneralForwardOut) -> GeneralLoss:
        raise NotImplementedError



import random
from typing import Union, Collection, List, Callable, Optional

import numpy as np
import torch
from torch import Tensor
from torch.utils.data import Sampler

from fandak.utils.misc import is_listy


def set_seed(seed: int = 1, fully_deterministic: bool = False):
    """
    Sets the python, numpy and pytorch's ransom seeds.
    But this doesn't mean that you are fully deterministic.
    If you set fully_deterministic to True, your code will run slower, more deterministic, but not fully deterministic.
    You should also care about num_workers in your code and set them to 0.
    Also see: https://pytorch.org/docs/stable/notes/randomness.html
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if fully_deterministic:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def change_multiprocess_strategy(strategy: str = "file_system"):
    """
    You might want to use this because of the following bug:
    https://github.com/pytorch/pytorch/issues/973
    """
    torch.multiprocessing.set_sharing_strategy(strategy)


def send_to_device(
    items: Collection[Tensor], device: torch.device
) -> Collection[Tensor]:
    """
    Send a single for a tuple of tensors
    """
    return tuple(map(lambda x: x.to(device), items))


def tensor_to_numpy(
    x: Union[Tensor, Collection[Tensor]]
) -> Union[np.ndarray, Collection[np.ndarray]]:
    if is_listy(x):
        return tuple(map(lambda t: tensor_to_numpy(t), x))
    else:
        return x.detach().cpu().numpy()


class OverfitSampler(Sampler):
    def __init__(self, main_source, indices):
        super().__init__(main_source)
        self.main_source = main_source
        self.indices = indices

        main_source_len = len(self.main_source)

        how_many = int(round(main_source_len / len(self.indices)))
        self.to_iter_from = []
        for _ in range(how_many):
            self.to_iter_from.extend(self.indices)

    def __iter__(self):
        return iter(self.to_iter_from)

    def __len__(self):
        return len(self.main_source)


class GeneralDataClass:
    def filter_attributes(
        self,
        the_filter: Callable[["GeneralDataClass", str], bool],
        initial_attr_list: Optional[Collection[str]] = None,
    ) -> List[str]:
        if initial_attr_list is None:
            initial_attr_list = dir(self)
        return [a for a in initial_attr_list if the_filter(self, a)]

    def get_attribute_names(self) -> List[str]:
        return self.filter_attributes(
            lambda gdc, a: not a.startswith("__") and not callable(getattr(gdc, a))
        )

    def to(self, device: torch.device):
        for attr_name in self.get_attribute_names():
            attr = getattr(self, attr_name)
            if isinstance(attr, torch.Tensor):
                setattr(self, attr_name, attr.to(device))

    def item(self):
        """
        This is like calling `.item()` for all of the attributes and setting their values to it.
        Will raise error if more than a single item is inside the tensor.
        """
        for attr_name in self.get_attribute_names():
            attr = getattr(self, attr_name)
            if isinstance(attr, torch.Tensor):
                setattr(self, attr_name, attr.item())



import datetime
import sys
from abc import ABC
from pathlib import Path
from pickle import dump as pdump
from typing import Optional, Tuple, List, Union, Iterable, Dict

import matplotlib.pylab as plt
import torch
from torch.nn.utils import clip_grad_norm_
from torch.optim.lr_scheduler import MultiStepLR, StepLR, LambdaLR
from torch.optim.optimizer import Optimizer
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm
from yacs.config import CfgNode

from fandak.core.datasets import Dataset, GeneralBatch
from fandak.core.evaluators import Evaluator, GeneralEvaluatorResult
from fandak.core.models import Model, GeneralLoss, GeneralForwardOut
from fandak.utils.metrics import ScalarMetricCollection
from fandak.utils.misc import get_git_commit_hash, print_with_time

RUN_INFO_TEMPLATE = """Time: {time}
Command: {command}
Git hash: {hash}
-----------------------------------------
{config}
"""

RUN_EXTRA_TEMPLATE = """-----------------------------------------
{extra}
"""

TORCH_EXT = "trc"
PICKLE_EXT = "pkl"
MODEL_FILE_NAME = "model"
OPTIMIZER_FILE_NAME = "optimizer"
SCHEDULER_FILE_NAME = "scheduler"

Scheduler = Union[MultiStepLR, StepLR, LambdaLR]


class Trainer(ABC):
    main_loss_metric_name = "main/loss"
    eval_metric_name_format = "eval/{}"

    def __init__(
        self,
        cfg: CfgNode,
        exp_name: str,
        train_db: Dataset,
        model: Model,
        device: torch.device = torch.device("cpu"),
        evaluators: Optional[Union[Iterable[Evaluator], Evaluator]] = None,
    ):
        self.cfg = cfg
        self.exp_name = exp_name
        self.train_db = train_db
        self.model = model
        self.device = device
        if evaluators is not None:
            if not isinstance(evaluators, Iterable):
                evaluators = [evaluators]
        else:
            evaluators = []
        self.evaluators = evaluators  # type: Iterable[Evaluator]

        self.save_every = 1
        self.eval_every = 1

        self.root = self.figure_root()
        self.optimizer = self.figure_optimizer()
        self.scheduler = self.figure_scheduler()
        self.clip_grad_norm = self.figure_clip_grad_norm()

        self.tb_root = self.root / "TB"
        self.model.to(device)
        self.epoch_num = 0
        self.iter_num = 0

        self.experiment_folder = self.root / self.exp_name
        self.experiment_folder.mkdir(exist_ok=True, parents=True)
        self.run_number = self._figure_run_number()
        self._set_run_folder()
        self.metrics = self.create_metrics()
        self.update_trainer_using_config()

    def _set_run_folder(self):
        self.run_folder = self.experiment_folder / str(self.run_number)
        self.tb_folder = self.tb_root / self.exp_name / str(self.run_number)
        self.writer = SummaryWriter(str(self.tb_folder))

    # noinspection PyMethodMayBeStatic
    def extra_info_for_run(self) -> Optional[dict]:
        return None

    def _save_info_of_run(self):
        extra_info = self.extra_info_for_run()

        config_dump = self.cfg.dump()

        final_value = RUN_INFO_TEMPLATE

        result = final_value.format(
            time=str(datetime.datetime.now()),
            command=" ".join(sys.argv),
            hash=get_git_commit_hash(),
            config=config_dump,
        )

        if extra_info:
            result += RUN_EXTRA_TEMPLATE.format(extra=extra_info)

        with open(self.run_folder / Path("info.txt"), "w") as f:
            f.write(result)

        # fixme: there is still a lot of problems with viewing here
        result_markdown_linebreak_fixed = result.replace("\n", "  \n")

        # Add the info to tensorboard for easier observation.
        self.writer.add_text("info.txt", result_markdown_linebreak_fixed)

    def save_run_report(
        self, report: str, name: str = "report", extension: str = "txt"
    ):
        with open(self.run_folder / Path("%s.%s" % (name, extension)), "w") as f:
            f.write(report)

    def save_fig(self, fig: plt.Figure, name: str = "figure"):
        fig.savefig(self.run_folder / ("%s.png" % name))

    def _mark_the_run(self):
        self.run_folder.mkdir(exist_ok=True, parents=True)
        self._save_info_of_run()

    def train(self):
        num_epochs = self.figure_num_epochs()
        self._mark_the_run()
        print_with_time("Training for run number: {:d}".format(self.run_number))
        epoch_range_start = self.epoch_num
        epoch_range = range(epoch_range_start, epoch_range_start + num_epochs)

        # callback
        self.on_start_training(num_epochs)

        for epoch_num in epoch_range:
            self.epoch_num = epoch_num

            # callback
            self.on_start_epoch(epoch_num)

            # resetting metrics
            for n, m in self.metrics.items():
                if self.metrics[n].report_average:
                    m.reset_values()

            # scheduler
            if self.scheduler is not None:
                self.scheduler.step(epoch_num)

            # training for 1 epoch
            dataloader = self.create_train_dataloader()
            self.train_1_epoch(epoch_num, dataloader)

            # saving
            if (epoch_num + 1) % self.save_every == 0:
                self.save_training()

            # evaluation
            if (epoch_num + 1) % self.eval_every == 0:
                eval_results = []
                for evaluator in self.evaluators:
                    eval_results.append(evaluator.evaluate())
                self.track_end_of_epoch_metrics(eval_results, epoch_num)

            # callback
            self.on_finish_epoch(epoch_num)

        # callback
        self.on_finish_training(num_epochs)

    def train_1_epoch(self, epoch_number: int, dataloader: DataLoader):
        print_with_time("Training epoch %d ..." % (epoch_number + 1))
        self.model.to(self.device)
        self.model.train()

        for batch in tqdm(dataloader):
            # callback
            self.on_start_batch(self.iter_num, batch)

            # train for 1 batch
            batch_loss, forward_out = self._train_1_batch(self.iter_num, batch)

            # update metrics for 1 batch
            self.track_training_metrics(batch, forward_out, batch_loss, self.iter_num)

            # call back
            self.on_finish_batch(self.iter_num, batch, forward_out, batch_loss)
            self.iter_num += 1

    # noinspection PyUnusedLocal
    def _train_1_batch(
        self, iter_num: int, batch: GeneralBatch
    ) -> Tuple[GeneralLoss, GeneralForwardOut]:
        # callback
        self.on_start_batch(self.iter_num, batch)

        # initial setup
        self.optimizer.zero_grad()
        batch.to(self.device)

        # forward pass
        forward_out = self.model.forward(batch)
        loss = self.model.loss(batch, forward_out)

        # backward pass
        loss.main.backward()

        # optional gradient clipping
        if self.clip_grad_norm is not None:
            clip_grad_norm_(self.model.parameters(), max_norm=self.clip_grad_norm)

        # optimizer step
        self.optimizer.step()

        # callback
        self.on_finish_batch(self.iter_num, batch, forward_out, loss)

        return loss, forward_out

    def save_training(self):
        print_with_time("Saving model ...")
        self.checkpoint_this(MODEL_FILE_NAME, self.model.state_dict(), torch_save=True)
        self.checkpoint_this(
            OPTIMIZER_FILE_NAME, self.optimizer.state_dict(), torch_save=True
        )
        if self.scheduler is not None:
            self.checkpoint_this(SCHEDULER_FILE_NAME, self.scheduler, torch_save=True)

    def load_training(self, run: int, epoch: int):
        """
        todo: make epoch optional by looking for the largest value for it.
        """
        assert epoch >= 0
        self.run_number = run
        self.epoch_num = epoch - 1  # todo: is this correct?

        # this is not a good thing to do, we need a way to save the iter_num on disk maybe.
        self.iter_num = (1 + self.epoch_num) * len(self.train_db)

        self._set_run_folder()
        check_pointing_folder = self._get_checkpointing_folder()

        model_file_name = check_pointing_folder / (
            "%s.%s" % (MODEL_FILE_NAME, TORCH_EXT)
        )
        self.model.load_state_dict(torch.load(model_file_name))

        optimizer_file_name = check_pointing_folder / (
            "%s.%s" % (OPTIMIZER_FILE_NAME, TORCH_EXT)
        )
        self.optimizer.load_state_dict(torch.load(optimizer_file_name))

        if self.scheduler is not None:
            scheduler_file_name = check_pointing_folder / (
                "%s.%s" % (SCHEDULER_FILE_NAME, TORCH_EXT)
            )
            self.scheduler = torch.load(scheduler_file_name)

    def _get_checkpointing_folder(self) -> Path:
        return self.run_folder / str(self.epoch_num + 1)

    def checkpoint_this(self, name: str, thing, torch_save: bool = False):
        check_pointing_folder = self._get_checkpointing_folder()
        check_pointing_folder.mkdir(exist_ok=True, parents=True)
        if torch_save:
            file_name = check_pointing_folder / Path("%s.%s" % (name, TORCH_EXT))
            torch.save(thing, file_name)
        else:
            file_name = check_pointing_folder / Path("%s.%s" % (name, PICKLE_EXT))
            with open(file_name, "wb") as f:
                pdump(thing, f)

    def on_start_training(self, num_epochs: int):
        pass

    def on_start_epoch(self, epoch_num: int):
        pass

    def on_start_batch(self, iter_num: int, batch: GeneralBatch):
        pass

    def on_finish_batch(
        self,
        iter_num: int,
        batch: GeneralBatch,
        forward_out: GeneralForwardOut,
        loss: GeneralLoss,
    ):
        pass

    def on_finish_epoch(self, epoch_num: int):
        pass

    def on_finish_training(self, num_epochs: int):
        pass

    def _figure_run_number(self) -> int:
        # fixme: this is not thread safe!
        max_run = 0
        for f in self.experiment_folder.iterdir():
            if f.is_dir():
                try:
                    f = int(str(f.name))
                except ValueError:
                    continue
                if f > max_run:
                    max_run = f

        return max_run + 1

    def figure_root(self) -> Path:
        raise NotImplementedError

    def figure_optimizer(self) -> Optimizer:
        raise NotImplementedError

    def figure_scheduler(self) -> Optional[Scheduler]:
        raise NotImplementedError

    def figure_num_epochs(self) -> int:
        raise NotImplementedError

    def create_train_dataloader(self) -> DataLoader:
        """
        If you want to overfit your model, you have to implement it here.
        """
        raise NotImplementedError

    def update_trainer_using_config(self):
        """
        This is called at the end of init.
        So change the implementation if you like.
        """
        pass

    # noinspection PyMethodMayBeStatic
    def figure_clip_grad_norm(self) -> Optional[float]:
        """
        TODO: Think about removing this and adding it as an optional hook.
        Return a positive value if you want gradient clipping.
        Return None is you don't want gradient clipping.
        """
        return None

    def create_metrics(self) -> Dict[str, ScalarMetricCollection]:
        """
        By default we create a report average metric for the main loss.
        Also  a non-report average metric for each evaluator.
        """
        default_loss_metric = ScalarMetricCollection(self.writer, "loss")
        metrics = {self.main_loss_metric_name: default_loss_metric}

        for i, evaluator in enumerate(self.evaluators):
            metrics[
                self.eval_metric_name_format.format(i + 1)
            ] = ScalarMetricCollection(
                self.writer, evaluator.get_name(), report_average=False
            )

        return metrics

    # noinspection PyUnusedLocal
    def track_training_metrics(
        self,
        batch: GeneralBatch,
        forward_out: GeneralForwardOut,
        loss: GeneralLoss,
        iter_num: int,
    ):
        """
        By default we will update the loss.
        """
        self.metrics[self.main_loss_metric_name].add_value(value=loss, step=iter_num)

    def track_end_of_epoch_metrics(
        self, eval_results: List[GeneralEvaluatorResult], epoch_num: int
    ):
        """
        By default we will update the loss for average reporting. Also if any evaluator is provided
        they will also be updated.
        """
        # for the main loss
        self.metrics[self.main_loss_metric_name].epoch_finished(epoch_num=epoch_num)

        # for each evaluator
        for i, (evaluator, eval_result) in enumerate(
            zip(self.evaluators, eval_results)
        ):
            # noinspection PyTypeChecker
            self.metrics[self.eval_metric_name_format.format(i + 1)].add_value(
                value=eval_result, step=(epoch_num + 1)
            )



from dataclasses import dataclass
from typing import List

from torch import Tensor
from torch.utils.data import Dataset as tDataset

# noinspection PyUnresolvedReferences
from torch.utils.data.dataloader import default_collate
from yacs.config import CfgNode

from fandak.utils.torch import GeneralDataClass


@dataclass
class GeneralBatch(GeneralDataClass):
    @classmethod
    def default_collate(cls, items: List["GeneralBatch"]) -> "GeneralBatch":
        """
        Uses default_collate from `torch.utils.data.Dataloader.default_collate`
        for every attribute of the type to collate them together.
        """
        values = {}
        for attr_name in items[0].get_attribute_names():
            if isinstance(getattr(items[0], attr_name), Tensor):
                attrs = [getattr(i, attr_name) for i in items]
                values[attr_name] = default_collate(attrs)

        # noinspection PyArgumentList
        return cls(**values)


class Dataset(tDataset):
    def __init__(self, cfg: CfgNode):
        super().__init__()
        self.cfg = cfg

    def __len__(self) -> int:
        raise NotImplementedError

    def __getitem__(self, item: int) -> GeneralBatch:
        raise NotImplementedError

    def collate_fn(self, items: List[GeneralBatch]) -> GeneralBatch:
        """
        Each dataset class should implement its own collate function.
        Could be as simple as `return GeneralBatch.default_collate(items)`
        """
        raise NotImplementedError



import datetime
import subprocess
from typing import Any


def get_git_commit_hash() -> str:
    """
    Returns the hash of the HEAD commit. Or an empty string.
    Can be used for logging purposes.
    """
    try:
        return (
            subprocess.check_output(["git", "rev-parse", "HEAD"])
            .decode("utf-8")
            .strip()
        )
    except subprocess.CalledProcessError:
        # this is probably not in a git repo or git is not installed.
        return ""


def print_with_time(the_thing: str):
    """
    Print an string with time.
    """
    print("[{}] {}".format(str(datetime.datetime.now()), the_thing))


def is_listy(x: Any) -> bool:
    """
    Grabbed this from fast.ai
    """
    return isinstance(x, (tuple, list))



from abc import ABC
from dataclasses import dataclass

import torch
from torch.utils.data import DataLoader
from tqdm import tqdm
from yacs.config import CfgNode

from fandak.core.datasets import Dataset, GeneralBatch
from fandak.core.models import Model, GeneralForwardOut
from fandak.utils.torch import GeneralDataClass


@dataclass
class GeneralEvaluatorResult(GeneralDataClass):
    pass


class Evaluator(ABC):
    def __init__(
        self,
        cfg: CfgNode,
        test_db: Dataset,
        model: Model,
        device: torch.device = torch.device("cpu"),
    ):
        self.cfg = cfg
        self.test_db = test_db
        self.model = model
        self.device = device

        # callback
        self.set_storage()

    def get_name(self) -> str:  # used in metrics
        return self.__class__.__name__

    def evaluate(self) -> GeneralEvaluatorResult:
        # callback
        self.on_start_eval()

        # prepare model
        self.model.to(self.device)
        self.model.eval()

        # create dataloader
        dataloader = self.create_dataloader()

        # perform evaluation
        with torch.no_grad():
            for batch in tqdm(dataloader):
                self._eval_1_batch(batch)

        # call back
        result = self.on_finish_eval()
        return result

    def _eval_1_batch(self, batch: GeneralBatch):
        batch.to(self.device)
        forward_out = self.model.forward(batch)
        self.batch_eval_calculation(batch, forward_out)

    def batch_eval_calculation(
        self, batch: GeneralBatch, forward_out: GeneralForwardOut
    ):
        raise NotImplementedError

    def create_dataloader(self) -> DataLoader:
        raise NotImplementedError

    def on_start_eval(self):
        pass

    def on_finish_eval(self) -> GeneralEvaluatorResult:
        raise NotImplementedError

    def set_storage(self):
        """
        create some storage to save intermediate evaluation results.
        """
        pass



from collections import defaultdict
from numbers import Number

from torch import Tensor
from torch.utils.tensorboard import SummaryWriter

from fandak.utils import print_with_time
from fandak.utils.torch import GeneralDataClass


class ScalarMetric:
    # TODO: refactor to inherit ScalarMetricCollection
    def __init__(self, writer: SummaryWriter, name: str, report_average: bool = True):
        self.writer = writer
        self.name = name
        self.report_average = report_average
        self.values = []
        self.average_tag = "training_average/%s" % self.name

    def add_value(self, value: float, step: int):
        self.writer.add_scalar(tag=self.name, scalar_value=value, global_step=step)
        self.values.append(value)

    def epoch_finished(self, epoch_num: int):
        average_value = self.average_value()
        if self.report_average:
            self.writer.add_scalar(
                tag=self.average_tag,
                scalar_value=average_value,
                global_step=epoch_num + 1,
            )
            print_with_time("%s: %f" % (self.average_tag, average_value))
        self.reset_values()

    def reset_values(self):
        self.values.clear()

    def average_value(self) -> float:
        return sum(self.values) / len(self.values)


class ScalarMetricCollection:
    def __init__(
        self, writer: SummaryWriter, base_name: str, report_average: bool = True
    ):
        self.writer = writer
        self.base_name = base_name
        self.report_average = report_average
        self.values = defaultdict(list)
        self.average_base_tag = "training_average/%s" % self.base_name

    def add_value(self, value: GeneralDataClass, step: int):
        def is_scalar_like(ds: GeneralDataClass, an: str) -> bool:
            a = getattr(ds, an)
            if isinstance(a, Tensor):  # is Tensor
                try:  # has .item()
                    a.item()
                except ValueError:
                    return False
                return True
            elif isinstance(a, Number):  # is Number
                return True
            else:
                return False

        loss_like_attr_names = value.filter_attributes(
            is_scalar_like, initial_attr_list=value.get_attribute_names()
        )

        for attr_name in loss_like_attr_names:
            tag_name = "{base_name}/{attr_name}".format(
                base_name=self.base_name, attr_name=attr_name
            )
            attr = getattr(value, attr_name)
            if isinstance(attr, Tensor):
                value = attr.item()
            else:
                value = attr
            self.writer.add_scalar(tag_name, scalar_value=value, global_step=step)
            self.values[attr_name].append(value)

    def epoch_finished(self, epoch_num: int):
        if self.report_average:
            for attr_name in self.values.keys():
                average_value = self.average_value(attr_name)
                tag_name = "{base_name}/{attr_name}".format(
                    base_name=self.average_base_tag, attr_name=attr_name
                )
                self.writer.add_scalar(
                    tag=tag_name, scalar_value=average_value, global_step=epoch_num + 1
                )
                print_with_time("%s: %f" % (tag_name, average_value))
        self.reset_values()

    def reset_values(self):
        for k in self.values.keys():
            self.values[k].clear()

    def average_value(self, attr_name: str) -> float:
        return sum(self.values[attr_name]) / len(self.values[attr_name])



import argparse

from yacs.config import CfgNode


def default_parse_args() -> argparse.Namespace:
    """
    Default parsing of the input arguments for a CLI. Assumes usage of YACS.
    a single optional config file is requested.
    then all other config files can be set by --set option.
    """
    parser = argparse.ArgumentParser(description="Train")
    parser.add_argument(
        "--cfg", dest="cfg_file", help="optional config file", default=None, type=str
    )
    parser.add_argument(
        "--set",
        dest="set_cfgs",
        help="set config keys",
        default=None,
        nargs=argparse.REMAINDER,
    )

    args = parser.parse_args()
    return args


def update_config(default_config: CfgNode) -> CfgNode:
    """
    This is useful for updating your config from CLI inputs, whether by a new config file or by --set.
    It will freeze the default config to prevent code smell.
    """
    args = default_parse_args()
    if args.cfg_file is not None:
        default_config.merge_from_file(args.cfg_file)
    if args.set_cfgs is not None:
        default_config.merge_from_list(args.set_cfgs)

    default_config.freeze()

    return default_config



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="fandak",
    version="0.0.3",
    author="Yasser Souri",
    author_email="yassersouri@gmail.com",
    description="A Framework for Deep Learning Research in PyTorch.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yassersouri/fandak",
    packages=setuptools.find_packages(exclude=['docs', 'examples', 'tests', 'scripts']),
    install_requires=[
        "torch>=1.0,<2",
        "torchvision>=0.3,<1",
        "tqdm>=4.32,<5",
        "matplotlib>=3.1.1,<4",
        "tensorboard>=1.14.0,<2",
        "future>=0.17.1,<1",
        "yacs>=0.1.6,<0.2",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Typing :: Typed",
    ],
)
